/*
 * i2s.c
 * mbedのI2Sクラスをマスタ＆送信専用で移植。
 *  Created on: 2018/05/26
 */

#ifdef __USE_CMSIS
#include "LPC17xx.h"
#endif

#include <math.h>		// fmod()
#include "i2s.h"

#define I2S_DF_WORDWIDTH                16
#define I2S_DF_SAMPLERATE               32000
#define I2S_DF_MASTERSLAVE              I2S_SLAVE
#define I2S_DF_STEREOMONO               I2S_STEREO
#define I2S_DF_MUTED                    I2S_UNMUTED
#define I2S_DF_INTERRUPT_FIFO_LEVEL     4

#define I2S_MAX_DENOMINATOR             256
#define I2S_MAX_NUMERATOR               256
#define I2S_MAX_BITRATE_DIV             64

#define I2S_PCLK_RATE                   24000000

static int mclk_frequency;
static int mclk_d;
//static int fourwire;		// 1=four pin mode
static int wordwidth;
static int mono;
static int freq;
static int master;
static int muted;
static int stopped;

static void i2s_pin_setup();
static void i2s_write_registers();
static void i2s_fraction_estimator(float in, int * num, int * den);
static int i2s_max_fifo_points();
static float i2s_mod(float in);

void i2s_init(){
	// I2S::I2S
    NVIC_DisableIRQ(I2S_IRQn);
    mclk_d = 0;
    mclk_frequency = 0;
    i2s_pin_setup();

    // 以下I2S::defaulter()から
    LPC_SC->PCONP |= (1 << 27);		// I2S interface power/clock control bit

    stopped = 1;
    freq = I2S_DF_SAMPLERATE;
    wordwidth = I2S_DF_WORDWIDTH;
    master = I2S_MASTER;
    mono = I2S_STEREO;
    //set_interrupt_fifo_level(I2S_DF_INTERRUPT_FIFO_LEVEL);
    muted = I2S_UNMUTED;
    i2s_write_registers();

    //NVIC_SetVector(I2S_IRQn, (uint32_t) & _i2sisr);
    //NVIC_EnableIRQ (I2S_IRQn);
}

static void i2s_pin_setup(){
	// 送信用ピン設定
    LPC_PINCON->PINSEL0 |= (0x01 << 18); //set P0.9 as transmit serial data line
    LPC_PINCON->PINSEL0 |= (0x01 << 14); //set P0.7 as transmit clock line
    LPC_PINCON->PINSEL0 |= (0x01 << 16); //set P0.8 as word select line
}

static void i2s_write_registers(){
    //Clock Multiplier Calculations
    float pre_mult = 0;
    int pre_num = 0;
    int pre_den = 0;
    int bitrate_div = 0;
    int bitrate = freq * 64;
    float target_div = I2S_PCLK_RATE / (float)(bitrate * 2);// Work out what divider is needed in the end, including the halving of rates the smoother does
    char wordwidth_code;

    if (mclk_frequency == 0) {
        float rnd = fmod(target_div, 1);// To make the X/Y fraction closest to 1, we set the last divider to the nearest integer to the rate divider
        bitrate_div = (int)(target_div - rnd);
        while (bitrate_div > I2S_MAX_BITRATE_DIV) { // But this might be out of range, so we right shift it into focus
            bitrate_div >>= 1;
        }
        if (bitrate_div == 0) { // Could be zero, which would disable the the clock...
            bitrate_div = 1;
        }
        pre_mult = (float)(bitrate_div) / target_div;    // Work out what we have left to correct
        pre_num = 0;
        pre_den = 0;
        i2s_fraction_estimator(pre_mult, &pre_num, &pre_den);// Get the function to work out the closest fraction, there might be some point in adding some possible multipliers of these values to add to the smoothing, the reference manual (UM10360 page 480) suggests this
    } else {
        pre_mult = (float)(mclk_frequency * 2) / (I2S_PCLK_RATE);
        pre_num = 0;
        pre_den = 0;
        i2s_fraction_estimator(pre_mult, &pre_num, &pre_den);// Get the function to work out the closest fraction, there might be some point in adding some possible multipliers of these values to add to the smoothing, the reference manual (UM10360 page 480) suggests this
        bitrate_div = (int)(
                          I2S_PCLK_RATE * (float)(pre_num) / (float)(pre_den)
                          / (float)(bitrate));
    }

    //Clock Multiplier, MCLK setup
    {
        int regvals = ((pre_num << 8) & 0xFF00) | (pre_den & 0xFF);
        LPC_I2S->I2STXRATE = regvals;                // Setting the X/Y fraction
        LPC_I2S->I2STXBITRATE = (bitrate_div - 1) & 0x3F;// Setting up the bitrate divider, the periferal adds one to this

        //LPC_I2S->I2STXMODE = fourwire << 2;

        if (mclk_d == 1) {
            LPC_I2S->I2STXMODE |= (1 << 3);		// Enable for the TX_MCLK output
        }
    }

    switch (wordwidth) {
        case 8:
            wordwidth_code = 0;
            break;
        case 16:
            wordwidth_code = 1;
            break;
        case 32:
            wordwidth_code = 3;
            break;
        default:
            //reg_write_err++;
            break;
    }

    int I2SDA_reg = (wordwidth_code & 0x3);
    I2SDA_reg |= ((mono << 2) & 0x4);		// BIT2:1=mono
    I2SDA_reg |= ((stopped << 3) & 0x8);		// BIT3:1=mute
    I2SDA_reg |= ((master << 5) & 0x20);	// BIT5:0=master
    I2SDA_reg |= (0x1F << 6);				// Word select half period minus 1,
    I2SDA_reg |= ((muted << 15) & 0x8000);	// BIT15:1=the transmit channel sends only zeroes

    LPC_I2S->I2SDAO = I2SDA_reg;

    LPC_I2S->I2SIRQ &= 0xFFFFFFFD;			// I2S transmit interrupt なし
}

// A function to find the nearest fraction to that put to it, with numerator and denomnator less than 256
// This is used when trying to get the clocks correct
static void i2s_fraction_estimator(float in, int * num, int * den){
    int test_num = 0;
    int test_den = 0;
    float least_error = 1;
    int least_err_den = 0;
    float genval;
    float generr;

    for (test_den = 1; test_den < I2S_MAX_DENOMINATOR; test_den++) {
        test_num = (int)((float)(test_den) * in);
        if (test_num < I2S_MAX_NUMERATOR && test_num > 0) {
            genval = (float)(test_num) / (float)(test_den);
            generr = i2s_mod(genval - in);
            if (generr < least_error) {
                least_error = generr;
                least_err_den = test_den;
            }
            if (generr == 0) {
                break;
            }
        }
    }

    test_num = (int)((float)(least_err_den) * in);
    *num = test_num;
    *den = least_err_den;

}

void i2s_frequency(int desired_freq){
    freq = desired_freq;
    i2s_write_registers();
}

void i2s_wordsize(int words){
    wordwidth = words;
    i2s_write_registers();
}

void i2s_masterslave(int mastermode){
	master = mastermode;
	i2s_write_registers();
}

void i2s_stereomono(int stereomono){
	mono = stereomono;
}

void i2s_mute(int mute_en){
    muted = mute_en;
    i2s_write_registers();
}

void i2s_start(){
    stopped = 0;
    muted = I2S_UNMUTED;
    i2s_write_registers();
}

static int i2s_max_fifo_points(){
    switch (wordwidth) {
        case 8:
            return (4 * 8);
        case 16:
            return (2 * 8);
        case 32:
            return 8;
        default:
            return 0;
    }
}

void i2s_write(unsigned short buf[], int len){
    //if (wordwidth > 0) {
        //if (len > i2s_max_fifo_points()) {
        //    len = i2s_max_fifo_points();
        //    //printf("Trying to write too much data!\n\r");
        //}
        //if (len <= 0)
        //    return;

        int i, j;
        uint32_t temp = 0;
        int increment = 32 / wordwidth;
        unsigned char recast[] = { 0, 0, 0, 0 };

        for (i = 0; i < len; i += increment) {
            temp = 0;

            switch (wordwidth) {
                case 8:
                    recast[0] = (uint8_t) buf[i + 0];
                    recast[1] = (uint8_t) buf[i + 1];
                    recast[2] = (uint8_t) buf[i + 2];
                    recast[3] = (uint8_t) buf[i + 3];
                    break;
                case 16:
                    recast[0] = (((uint16_t) buf[i + 0]) >> 0) & 0xFF;
                    recast[1] = (((uint16_t) buf[i + 0]) >> 8) & 0xFF;
                    recast[2] = (((uint16_t) buf[i + 1]) >> 0) & 0xFF;
                    recast[3] = (((uint16_t) buf[i + 1]) >> 8) & 0xFF;
                    break;
                case 32:
                    recast[0] = (((uint32_t) buf[i + 0]) >> 0) & 0xFF;
                    recast[1] = (((uint32_t) buf[i + 0]) >> 8) & 0xFF;
                    recast[2] = (((uint32_t) buf[i + 0]) >> 16) & 0xFF;
                    recast[3] = (((uint32_t) buf[i + 0]) >> 24) & 0xFF;
                    break;
            }

            for (j = 0; j < 4; j++) {
                temp |= recast[j] << (j * 8);
            }

            //if(((temp >> 16) & 0xFFFF) == 0xFFFF) printf("Hmmm %x %x %x\n\r",temp, increment,i); //|| temp &0xFFFF == 0xFFFF
            //if((buf[i]-buf[i+1])>5000 || (buf[i]-buf[i+1])<-5000) printf("J:%i,%i\n\r",buf[i],buf[i+1]);
            //printf("%x\n",temp);
            LPC_I2S->I2STXFIFO = temp;
        }
    //}
}

static float i2s_mod(float in){
    if (in < 0)
        in *= -1;

    return in;
}

void i2s_begin(int sample_freq, int wordsize){
	i2s_frequency(sample_freq);
	i2s_wordsize(wordsize);
	i2s_start();
}

void i2s_end(){
    stopped = 0;
    muted = I2S_MUTED;
    i2s_write_registers();
}
